-- This file contains variables or functions that are related to the Duplicating Chest family in this mod.
if not duplicating_chest_util then duplicating_chest_util = {} end

-- Returns the inventory of the given chest.
function duplicating_chest_util.get_inventory(chest)
	local inventory = chest.get_output_inventory()
	if not inventory then
		inventory = chest.get_inventory(defines.inventory.cargo_wagon)
	end
	return inventory
end

-- Refills some chests among the given array of chests according to the given index of the next chest to update.
-- Returns the index of the next chest to update in the next tick.
function duplicating_chest_util.duplicate_contents(chests, next_chest_to_update)
	if #chests <= 0 then
		return 1
	end
	
	local chest_index = next_chest_to_update
	for i = 1, 10, 1 do
		local chest = chests[chest_index]
		if chest.valid then
			if not chest.to_be_deconstructed(chest.force) then
				local inventory = duplicating_chest_util.get_inventory(chest)
				local item = inventory[1]
				-- item.valid is true as long as the entity is valid. So it cannot be used for checking whether there is actually an item stack.
				-- We have to use valid_for_read instead.
				if item ~= nil and item.valid_for_read then
					-- Fill the whole chest immediately.
					inventory.insert{name = item.name, count = item.prototype.stack_size * #inventory}
				end
			end
			
			-- Prepare for the next chest.
			chest_index = chest_index + 1
			-- No more next chest. Return to the first chest.
			if chest_index > #chests then
				return 1
			end
		else
			table.remove(chests, chest_index)
			if chest_index > #chests then
				return 1
			end
		end		
	end
	-- Return the index of the next chest to update.
	return chest_index
end

-- Copies the contents from the source chest to the destination chest when copy-paste is done on the Duplicating Chest family.
function duplicating_chest_util.on_entity_copied_pasted(source_chest, destination_chest)
	local source_inventory = duplicating_chest_util.get_inventory(source_chest)
	local destination_inventory = duplicating_chest_util.get_inventory(destination_chest)
	
	destination_inventory.clear()
	
	local source_item = source_inventory[1]
	if source_item ~= nil and source_item.valid_for_read then
		destination_inventory.insert{name = source_item.name, count = source_item.prototype.stack_size * #destination_inventory}
	end
end