-- This file contains variables or functions that are related to the Creator magic wand in this mod.
if not magic_wand_creator then magic_wand_creator = {} end

-- Returns whether tile correction should be performed for the given player.
function magic_wand_creator.get_tile_correction(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		if global.creative_mode.magic_wand_settings.creator[player.index].tile_correction ~= nil then
			return global.creative_mode.magic_wand_settings.creator[player.index].tile_correction
		end
	end
	return true
end

-- Sets whether tile correction should be performed for the given player.
function magic_wand_creator.set_tile_correction(player, tile_correction)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].tile_correction = tile_correction
end

----

-- Returns whether the don't-kill-players-by-tiles option is turned on for the given player.
function magic_wand_creator.get_dont_kill_players_by_tiles(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		if global.creative_mode.magic_wand_settings.creator[player.index].dont_kill_players_by_tiles ~= nil then
			return global.creative_mode.magic_wand_settings.creator[player.index].dont_kill_players_by_tiles
		end
	end
	return true
end

-- Sets whether the don't-kill-players-by-tiles option is turned on for the given player.
function magic_wand_creator.set_dont_kill_players_by_tiles(player, dont_kill_players_by_tiles)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].dont_kill_players_by_tiles = dont_kill_players_by_tiles
end

----

-- Returns the tile prototype selected by the given player.
-- If no tile is selected, nil will be returned.
function magic_wand_creator.get_selected_tile_prototype(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		local tile_name = global.creative_mode.magic_wand_settings.creator[player.index].tile_name
		if tile_name then
			return game.tile_prototypes[tile_name]
		end
	end
	-- Default to concrete. It can be nil if not existed.
	return game.tile_prototypes["concrete"]
end

-- Sets the given tile prototype as the selected tile by the given player. It can be nil.
function magic_wand_creator.set_selected_tile_prototype(player, tile_prototype)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	if tile_prototype then
		global.creative_mode.magic_wand_settings.creator[player.index].tile_name = tile_prototype.name
	else
		global.creative_mode.magic_wand_settings.creator[player.index].tile_name = ""
	end
	global.creative_mode.magic_wand_settings.creator[player.index].resource_name = "" 
end

----

-- Returns the resource entity prototype selected by the given player.
-- If no resource is selected, nil will be returned.
function magic_wand_creator.get_selected_resource_prototype(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		local resource_name = global.creative_mode.magic_wand_settings.creator[player.index].resource_name
		if resource_name then
			return game.entity_prototypes[resource_name]
		end
	end
	return nil
end

-- Sets the given resource entity prototype as the selected resource by the given player. It can be nil.
function magic_wand_creator.set_selected_resource_prototype(player, resource_prototype)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].tile_name = ""
	if resource_prototype then
		global.creative_mode.magic_wand_settings.creator[player.index].resource_name = resource_prototype.name
	else
		global.creative_mode.magic_wand_settings.creator[player.index].resource_name = ""
	end
end

----

-- Returns the resource amount when creating resources for the given player.
function magic_wand_creator.get_resource_amount(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		local resource_amount = global.creative_mode.magic_wand_settings.creator[player.index].resource_amount
		if resource_amount then
			return resource_amount
		end
	end
	return 2000
end

-- Sets the resource amount when creating resources for the given player.
function magic_wand_creator.set_resource_amount(player, resource_amount)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].resource_amount = resource_amount
end

----

-- Returns whether the don't-remove-player-characters option is turned on for the given player.
function magic_wand_creator.get_dont_remove_player_characters(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		if global.creative_mode.magic_wand_settings.creator[player.index].dont_remove_player_characters ~= nil then
			return global.creative_mode.magic_wand_settings.creator[player.index].dont_remove_player_characters
		end
	end
	return true
end

-- Sets whether the don't-remove-player-characters option is turned on for the given player.
function magic_wand_creator.set_dont_remove_player_characters(player, dont_remove_player_characters)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].dont_remove_player_characters = dont_remove_player_characters
end

----

-- Returns whether tile removal should not be performed when there is any entity being selected for the given player.
function magic_wand_creator.get_dont_remove_tiles_if_any_entity_is_selected(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		if global.creative_mode.magic_wand_settings.creator[player.index].dont_remove_tiles_if_any_entity_is_selected ~= nil then
			return global.creative_mode.magic_wand_settings.creator[player.index].dont_remove_tiles_if_any_entity_is_selected
		end
	end
	return true
end

-- Sets whether tile removal should not be performed when there is any entity being selected for the given player.
function magic_wand_creator.set_dont_remove_tiles_if_any_entity_is_selected(player, dont_remove_tiles_if_any_entity_is_selected)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].dont_remove_tiles_if_any_entity_is_selected = dont_remove_tiles_if_any_entity_is_selected
end

----

-- Returns whether the don't-kill-players-by-removing-tiles option is turned on for the given player.
function magic_wand_creator.get_dont_kill_players_by_removing_tiles(player)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		if global.creative_mode.magic_wand_settings.creator[player.index].dont_kill_players_by_removing_tiles ~= nil then
			return global.creative_mode.magic_wand_settings.creator[player.index].dont_kill_players_by_removing_tiles
		end
	end
	return true
end

-- Sets whether the don't-kill-players-by-removing-tiles option is turned on for the given player.
function magic_wand_creator.set_dont_kill_players_by_removing_tiles(player, dont_kill_players_by_removing_tiles)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].dont_kill_players_by_removing_tiles = dont_kill_players_by_removing_tiles
end

----

-- Returns whether the alt-mode of the magic wand can apply on the given force for the given player.
function magic_wand_creator.get_alt_mode_apply_on_force(player, force)
	if global.creative_mode.magic_wand_settings.creator[player.index] then
		if global.creative_mode.magic_wand_settings.creator[player.index].alt_mode_forces ~= nil then
			return global.creative_mode.magic_wand_settings.creator[player.index].alt_mode_forces[force.name] ~= false
		end
	end
	return true
end

-- Sets whether the alt-mode of the magic wand can apply on the given force for the given player.
function magic_wand_creator.set_alt_mode_apply_on_force(player, force, apply)
	if not global.creative_mode.magic_wand_settings.creator[player.index] then global.creative_mode.magic_wand_settings.creator[player.index] = {} end
	if not global.creative_mode.magic_wand_settings.creator[player.index].alt_mode_forces then global.creative_mode.magic_wand_settings.creator[player.index].alt_mode_forces = {} end
	global.creative_mode.magic_wand_settings.creator[player.index].alt_mode_forces[force.name] = apply
end

------------------------------------------------------

-- Creates smoke effect at the given position in the given surface.
local function create_smoke_at(surface, position)
	surface.create_entity
	{
		name = creative_mode_defines.names.entities.magic_wand_smoke_creator,
		position = position
	}
end

-- Creates smoke effect at the entity's position.
local function create_smoke_effect_at_entity_position(entity)
	create_smoke_at(entity.surface, entity.position)
end

------------------------------------------------------

-- Returns the dictionary of all safe positions for all connected players with characters on the given surface.
local function get_safe_positions_for_all_players_on_surface(surface)
	local player_positions
	for _, player in pairs(game.connected_players) do
		local character = player.character
		if character then
			if character.surface == surface then
				if not player_positions then player_positions = {} end
				local position = character.position
				local rounded_x = util.round(position.x)
				local rounded_y = util.round(position.y)
				
				-- Give him a 3x3 safe area.
				local x = rounded_x - 1
				if not player_positions[x] then player_positions[x] = {} end
				player_positions[x][rounded_y - 1] = true
				player_positions[x][rounded_y] = true
				player_positions[x][rounded_y + 1] = true
				
				x = rounded_x
				if not player_positions[x] then player_positions[x] = {} end
				player_positions[x][rounded_y - 1] = true
				player_positions[x][rounded_y] = true
				player_positions[x][rounded_y + 1] = true
				
				x = rounded_x + 1
				if not player_positions[x] then player_positions[x] = {} end
				player_positions[x][rounded_y - 1] = true
				player_positions[x][rounded_y] = true
				player_positions[x][rounded_y + 1] = true
			end
		end
	end
	return player_positions
end

-- Handler of the on_player_selected_area event.
-- Returns whether the event is consumed and hence no need to further pass it to other handlers.
function magic_wand_creator.on_player_selected_area(player, area, item_name, entities, tiles)
	if item_name == creative_mode_defines.names.items.magic_wand_creator then
		if rights.can_player_use_creator_magic_wand(player) then
			-- Create tiles.
			local surface = player.surface
			local selected_tile = magic_wand_creator.get_selected_tile_prototype(player)
			if selected_tile then
				local new_tile_name = selected_tile.name
				local dont_kill_players_by_tiles = magic_wand_creator.get_dont_kill_players_by_tiles(player) and selected_tile.collision_mask["player-layer"]
				local player_positions = nil
				if dont_kill_players_by_tiles then
					-- Don't kill players, but the selected tile is harmful.
					-- Make sure the positions of all players are not touched.
					player_positions = get_safe_positions_for_all_players_on_surface(surface)
				end
				local new_tiles
				local positions
				for _, tile in pairs(tiles) do
					if tile.name ~= new_tile_name then
						local position = tile.position
						if not player_positions or not player_positions[position.x] or not player_positions[position.x][position.y] then
							-- Prepare to set tiles.
							if not new_tiles then new_tiles = {} end
							if not positions then positions = {} end
							table.insert(new_tiles,
							{
								name = new_tile_name,
								position = position
							})
							table.insert(positions, position)
							create_smoke_at(surface, position)
						end
					end
				end
				if new_tiles then
					surface.set_tiles(new_tiles, magic_wand_creator.get_tile_correction(player))
					-- Raise event.
					util.raise_event(defines.events.on_player_built_tile,
					{
						player_index = player.index,
						positions = positions
					})
				end
			else
				-- Or create resources.
				local selected_resource = magic_wand_creator.get_selected_resource_prototype(player)
				local resource_amount = magic_wand_creator.get_resource_amount(player)
				if selected_resource then
					local new_resource_name = selected_resource.name
					for _, tile in pairs(tiles) do
						if surface.can_place_entity{name = new_resource_name, position = tile.position} then
							local entity = surface.create_entity
							{
								name = new_resource_name,
								position = tile.position,
								amount = resource_amount
							}
							if entity then
								create_smoke_effect_at_entity_position(entity)
								-- Raise event.
								util.raise_event(defines.events.on_built_entity,
								{
									created_entity = entity,
									player_index = player.index
								})
							end
						end
					end
				end
			end
		else
			-- No right to use.
			player.print{"message.creative-mode_no-right-to-use-magic-wand"}
		end
		
		return true
	end
	
	return false
end

-- Handler of the on_player_alt_selected_area event.
-- Returns whether the event is consumed and hence no need to further pass it to other handlers.
function magic_wand_creator.on_player_alt_selected_area(player, area, item_name, entities, tiles)
	if item_name == creative_mode_defines.names.items.magic_wand_creator then
		if rights.can_player_use_creator_magic_wand(player) then
			-- Removes the selected entities.
			local has_destroyed_entity = false
			local dont_remove_player_characters = magic_wand_creator.get_dont_remove_player_characters(player)
			for _, entity in ipairs(entities) do
				if entity.valid then -- It is possible that the entity becomes invalid.
					-- Don't select the smoke.
					if entity.name ~= creative_mode_defines.names.entities.magic_wand_smoke_creator then
						local force = entity.force
						if force then
							-- Force allowed?
							if magic_wand_creator.get_alt_mode_apply_on_force(player, force) then
								-- Player character.
								local can_destroy = true
								if dont_remove_player_characters then
									if util.get_character_owning_player(entity) then
										-- The character is owned by a player.
										can_destroy = false
									end
								end
								if can_destroy then
									create_smoke_effect_at_entity_position(entity)
									util.destroy_entity_and_raise_event(entity, player, false)
									has_destroyed_entity = true
								end
							end
						end
					end
				end
			end
			
			-- Also remove tiles.
			-- But check if we can do so first.
			if has_destroyed_entity and magic_wand_creator.get_dont_remove_tiles_if_any_entity_is_selected(player) then
				-- No, don't remove tiles.
				return
			end			
			local surface = player.surface
			local new_tiles = nil
			local positions = nil
			local player_positions = nil
			local dont_kill_players_by_removing_tiles = magic_wand_creator.get_dont_kill_players_by_removing_tiles(player)
			for _, tile in ipairs(tiles) do
				local position = tile.position
				-- Remove the tile if it is hidding a tile behind it.
				local hidden_tile_name = surface.get_hidden_tile(position)
				local hidden_tile = game.tile_prototypes[hidden_tile_name]
				if hidden_tile then
					local remove_this_tile = true
					if hidden_tile.collision_mask["player-layer"] and dont_kill_players_by_removing_tiles then
						-- The hidden tile is unsafe for players.
						-- Make sure the positions of all players are not touched.
						if not player_positions then player_positions = get_safe_positions_for_all_players_on_surface(surface) end
						if player_positions[position.x] and player_positions[position.x][position.y] then
							remove_this_tile = false
						end
					end
					if remove_this_tile then
						if not new_tiles then new_tiles = {} end
						if not positions then positions = {} end
						table.insert(new_tiles,
						{
							name = hidden_tile_name,
							position = position
						})
						table.insert(positions, position)
						create_smoke_at(surface, position)	
					end
				end
			end
			if new_tiles then
				surface.set_tiles(new_tiles)
				-- Raise event.
				util.raise_event_for_destroyed_tiles(positions, player.index, false)
			end
		else
			-- No right to use.
			player.print{"message.creative-mode_no-right-to-use-magic-wand"}
		end
		
		return true
	end
	
	return false
end